import os
import itertools
import numpy as np
import matplotlib.pyplot as plt

from sklearn.preprocessing import MinMaxScaler
from sklearn.datasets import make_regression

from sklearn.inspection import DecisionBoundaryDisplay
import matplotlib.font_manager as fm

myfont = fm.FontProperties(fname="./FiraSans-Regular.ttf")


def sqdists(x, y):
    return np.sum((x[:, None, :] - y[None, :, :]) ** 2, axis=-1)


def kernel_matrix(x, y, *, kernel, gamma):
    if kernel == "cauchy":
        return 1 / (1 + gamma * sqdists(x, y))
    elif kernel == "rbf":
        return np.exp(-gamma * sqdists(x, y))
    elif kernel == "laplace":
        return np.exp(-gamma * np.sqrt(sqdists(x, y)))
    elif kernel == "exponential":
        return np.exp(gamma * x @ y.T)
    elif kernel == "distance" and gamma == 1:
        return -np.sqrt(sqdists(x, y))
    elif kernel == "distance" and gamma == -1 / 2:
        return 1 / (np.sqrt(np.sqrt(sqdists(x, y))) + 1e-8)
    elif kernel == "distance" and gamma == -1:
        return 1 / (np.sqrt(sqdists(x, y)) + 1e-8)
    elif kernel == "distance" and gamma == -2:
        return 1 / (sqdists(x, y) + 1e-8)
    elif kernel == "distance" and gamma == -4:
        return 1 / (sqdists(x, y) ** 2 + 1e-8)


class NWKernel:
    def __init__(self, *, kernel, gamma=None):
        self.kernel = kernel
        self.gamma = gamma

    def fit(self, X, y):
        self.X_ = X
        self.y_ = y

    def predict(self, X):
        K = kernel_matrix(X, self.X_, kernel=self.kernel, gamma=self.gamma)
        return (K @ self.y_) / np.sum(K, axis=1)


models = [
    # Raw dataset:
    ("dataset", None),
    ("attention_1", NWKernel(kernel="exponential", gamma=1)),
    ("attention_2", NWKernel(kernel="exponential", gamma=2)),
    ("attention_5", NWKernel(kernel="exponential", gamma=5)),
    ("attention_10", NWKernel(kernel="exponential", gamma=10)),
    ("attention_20", NWKernel(kernel="exponential", gamma=20)),
    ("attention_50", NWKernel(kernel="exponential", gamma=50)),
    ("attention_100", NWKernel(kernel="exponential", gamma=100)),
    ("attention_200", NWKernel(kernel="exponential", gamma=200)),
    ("attention_500", NWKernel(kernel="exponential", gamma=500)),
    ("attention_1000", NWKernel(kernel="exponential", gamma=1000)),
]


datasets = [
    ("2D_n100", make_regression(n_samples=20, n_features=2, noise=100, random_state=0)),
]

datasets = [
    (
        ds_name,
        (
            MinMaxScaler((-0.4, 0.4)).fit_transform(X),
            MinMaxScaler((-1, 1)).fit_transform(y.reshape(-1, 1)).reshape(-1),
        ),
    )
    for ds_name, (X, y) in datasets
]


X_1D = np.array([1.0, 1.4, 2.0, 2.3, 2.9, 4.0, 4.5, 5.0, 5.4]).reshape(-1, 1) / 6 - 0.5
y_1D = np.array([1.5, 0.8, 0.7, 0.2, 0.42, 1.0, 1.4, 1.1, 1.9]) / 6 - 0.25

datasets = [
    ("1D_bowl", (X_1D, y_1D)),
] + datasets

print(datasets)


# iterate over datasets

for ds_name, (X, y) in datasets:
    print(ds_name, "=====================================")
    D = X.shape[1]

    if D == 1:
        figure = plt.figure(figsize=(8, 5), dpi=300, facecolor="#fafafa")

    else:
        figure = plt.figure(figsize=(12, 12), facecolor="#fafafa")
        cm = plt.cm.RdBu
        vmin = -1.2
        vmax = 1.2

    for name, model in models:
        print(name, end=": ", flush=True)
        if model is not None:
            model.fit(X, y)
            pred = model.predict(X)
            print(np.mean(np.abs(pred - y)))

        plt.clf()
        ax = plt.gca()
        ax.set_facecolor("#fafafa")

        if D == 1:
            T = np.linspace(-0.5, 0.5, 10001).reshape(-1, 1)
            if model is not None:
                u = model.predict(T)

                segs_x = [[a, a, None] for a in X.reshape(-1)]
                segs_x = np.array(list(itertools.chain.from_iterable(segs_x)))
                segs_y = [[b, c, None] for b, c in zip(y, pred)]
                segs_y = np.array(list(itertools.chain.from_iterable(segs_y)))

            # get width and height of axes object to compute
            # matching arrowhead length and width
            dps = plt.gcf().dpi_scale_trans.inverted()
            bbox = ax.get_window_extent().transformed(dps)
            width, height = bbox.width, bbox.height

            xmin, xmax = -0.5, 0.5
            ymin, ymax = -0.3, 0.2
            # manual arrowhead width and length
            hw = 1.0 / 20.0 * (ymax - ymin)
            hl = 1.0 / 20.0 * (xmax - xmin)
            lw = 1.0  # axis line width
            ohg = 0.3  # arrow overhang

            # compute matching arrowhead length and width
            yhw = hw / (ymax - ymin) * (xmax - xmin) * height / width
            yhl = hl / (xmax - xmin) * (ymax - ymin) * width / height

            # draw x and y axis
            ax.arrow(
                xmin,
                -0.25,
                xmax - xmin,
                0.0,
                fc="k",
                ec="k",
                lw=lw,
                head_width=hw,
                head_length=hl,
                overhang=ohg,
                length_includes_head=True,
                clip_on=False,
            )

            ax.arrow(
                -0.45,
                ymin,
                0.0,
                ymax - ymin,
                fc="k",
                ec="k",
                lw=lw,
                head_width=.8*yhw,
                head_length=yhl,
                overhang=ohg,
                length_includes_head=True,
                clip_on=False,
            )

            ax.text(
                0.45,
                -0.3,
                r"x",
                fontproperties=myfont,
                fontsize=35,
                color="#080899",
            )
            ax.text(
                -0.495,
                0.15,
                r"y",
                fontproperties=myfont,
                fontsize=35,
                color="#005F5F", #"#990808",
            )

            if model is not None:
                ax.scatter(
                    X.reshape(-1),
                    pred,
                    40, #160,
                    color="b",
                    linewidths=1,
                    edgecolors="#000080",
                )
                ax.plot(T.reshape(-1), u, color="b", linewidth=2)
                ax.plot(segs_x, segs_y, color="#95DFEB", linewidth=4, zorder=0)

            ax.scatter(
                X.reshape(-1), y, 160, color="c", linewidths=2, edgecolors="#005F5F", zorder=4
            )

            ax.axis([-0.5, 0.5, -0.3, 0.2])
            ax.set_aspect("equal")
            ax.axis("off")

        elif D == 2:
            if model is not None:
                DecisionBoundaryDisplay.from_estimator(
                    model,
                    np.array([[-0.5, -0.5], [0.5, 0.5]]),
                    cmap=cm,
                    alpha=0.8,
                    ax=ax,
                    eps=0,
                    plot_method="pcolormesh",
                    grid_resolution=1000,
                    vmin=vmin,
                    vmax=vmax,
                )


            Q = make_regression(n_samples=20, n_features=2, noise=100, random_state=2)[0]
            Q = MinMaxScaler((-0.4, 0.4)).fit_transform(Q)

            ax.scatter(
                Q[:, 0],
                Q[:, 1],
                s=16**2,
                c="k",
                marker="x",
                cmap=cm,
                linewidths=3,
                vmin=vmin,
                vmax=vmax,
            )


            # Plot the training points
            ax.scatter(
                X[:, 0],
                X[:, 1],
                s=25**2,
                c=y,
                cmap=cm,
                linewidths=5,
                edgecolors="k",
                vmin=vmin,
                vmax=vmax,
            )


            ax.set_xlim(-0.5, 0.5)
            ax.set_ylim(-0.5, 0.5)
            ax.set_xticks(())
            ax.set_yticks(())

        plt.tight_layout()
        plt.show()

        filename = f"output_regression/{ds_name}/{name}.png"
        os.makedirs(os.path.dirname(filename), exist_ok=True)
        plt.savefig(filename, bbox_inches="tight", transparent=False)
